<!DOCTYPE html
    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:py="http://genshi.edgewall.org/"
      xmlns:xi="http://www.w3.org/2001/XInclude">
  <xi:include href="layout.html" />
  <head>
    <title>Upload images for gadget borders :</title>
  </head>

  <body>
    <div id="content" class="wiki">

      <h1>Create an image-based gadget border </h1>

      <img src="../../../chrome/ig/icon_home1.png" style="float:left;margin-right:20px;"/>
      <div id="description" xml:space="preserve">
        Select the images you want to show around
        <a class="ext-link" href="http://www.google.com/ig/">
          <span class="icon">iGoogle</span>
        </a>
        <a class="ext-link" href="http://code.google.com/apis/gadgets/">
          <span class="icon">Gadgets</span>
        </a>
        in your pages once this border is specified.
      </div>

      <div class="buttons" style="float:clear;">
        <form class="mod" action="" method="post" enctype="multipart/form-data" >
          <input type="hidden" name="action" value="new" />
          <fieldset>
            <legend>Border properties</legend>
            <div class="field">
              <label>Specify a name to identify this border:
                <input type="text" id="id" name="id" size="40" value="${fields and fields['id']}"/> <br />
              </label>
            </div>
            <hr />
            <h3>Select images :</h3>
            <div style="text-align: center">
              <div py:for="i, img_spec in enumerate([{'fnm' : 'b', 'w':1, 'h':28, 'loc':_('bottom')},
                    {'fnm' : 'tt', 'w':1, 'h':28, 'loc':_('top (with title)')},
                    {'fnm' : 'tn', 'w':1, 'h':12, 'loc':_('top (no title)')},
                    {'fnm' : 'l', 'w':8, 'h':1, 'loc':_('left side')},
                    {'fnm' : 'r', 'w':8, 'h':1, 'loc':_('right side')},
                    {'fnm' : 'bl', 'w':8, 'h':28, 'loc':_('bottom-left corner')},
                    {'fnm' : 'br', 'w':8, 'h':28, 'loc':_('bottom-right corner')},
                    {'fnm' : 'tl', 'w':8, 'h':28, 'loc':_('top-left corner')},
                    {'fnm' : 'tr', 'w':8, 'h':28, 'loc':_('top-right corner')}])"
                    class="field" style="float:left; margin-right: 10px;text-align: left;">
                <label>
                  Image at the ${img_spec['loc']}: <br />
                  <input type="file" id="${img_spec['fnm']}" name="${img_spec['fnm']}" />
                </label>
              </div>
            </div>
          </fieldset>
          <div class="buttons">
            <input type="submit" value="Upload" accesskey="u" />
          </div>
        </form>
      </div>
    </div>
  </body>
</html>
